/*
 * arith.c -- 64 bit arithmetic
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "arith.h"


octa zero = { 0, 0 };
octa one = { 0, 1 };
octa negone = { -1, -1 };

octa arithAuxOut;
bool arithOvrflw;


octa incrOcta(octa x, int delta) {
  octa r;

  r.h = x.h;
  r.l = x.l + delta;
  if (delta >= 0) {
    if (r.l < x.l) {
      r.h++;
    }
  } else {
    if (r.l > x.l) {
      r.h--;
    }
  }
  return r;
}


octa addOcta(octa x, octa y) {
  octa r;

  r.h = x.h + y.h;
  r.l = x.l + y.l;
  if (r.l < x.l) {
    r.h++;
  }
  return r;
}


octa subOcta(octa x, octa y) {
  octa r;

  r.h = x.h - y.h;
  r.l = x.l - y.l;
  if (r.l > x.l) {
    r.h--;
  }
  return r;
}


octa umulOcta(octa x, octa y) {
  tetra u[4], v[4], w[8];
  tetra t;
  int k;
  octa accumulator;
  int j, i;

  /* unpack multiplier and multiplicand */
  u[3] = x.h >> 16;
  u[2] = x.h & 0xFFFF;
  u[1] = x.l >> 16;
  u[0] = x.l & 0xFFFF;
  v[3] = y.h >> 16;
  v[2] = y.h & 0xFFFF;
  v[1] = y.l >> 16;
  v[0] = y.l & 0xFFFF;
  /* reset lower half of product */
  for (j = 0; j < 4; j++) {
    w[j] = 0;
  }
  /* compute product */
  for (j = 0; j < 4; j++) {
    if (v[j] == 0) {
      w[j + 4] = 0;
    } else {
      k = 0;
      for (i = 0; i < 4; i++) {
        t = u[i] * v[j] + w[i + j] + k;
        w[i + j] = t & 0xFFFF;
        k = t >> 16;
      }
      w[j + 4] = k;
    }
  }
  /* pack product into outputs */
  arithAuxOut.h = (w[7] << 16) + w[6];
  arithAuxOut.l = (w[5] << 16) + w[4];
  accumulator.h = (w[3] << 16) + w[2];
  accumulator.l = (w[1] << 16) + w[0];
  return accumulator;
}


octa smulOcta(octa x, octa y) {
  octa accumulator;

  accumulator = umulOcta(x, y);
  if ((x.h & MSB(32)) != 0) {
    arithAuxOut = subOcta(arithAuxOut, y);
  }
  if ((y.h & MSB(32)) != 0) {
    arithAuxOut = subOcta(arithAuxOut, x);
  }
  arithOvrflw =
    arithAuxOut.h != arithAuxOut.l ||
    (arithAuxOut.h ^ (arithAuxOut.h >> 1) ^ (accumulator.h & MSB(32))) != 0;
  return accumulator;
}


octa udivOcta(octa x, octa y, octa z) {
  tetra u[8], v[4], q[4];
  int n;
  tetra vh, vmh;
  int d;
  tetra t;
  int k;
  tetra qhat, rhat;
  tetra mask;
  octa accumulator;
  int j, i;

  /* check that x < z, otherwise give trivial answer */
  if (x.h > z.h || (x.h == z.h && x.l >= z.l)) {
    arithAuxOut = y;
    return x;
  }
  /* unpack the dividend and divisor to u and v */
  u[7] = x.h >> 16;
  u[6] = x.h & 0xFFFF;
  u[5] = x.l >> 16;
  u[4] = x.l & 0xFFFF;
  u[3] = y.h >> 16;
  u[2] = y.h & 0xFFFF;
  u[1] = y.l >> 16;
  u[0] = y.l & 0xFFFF;
  v[3] = z.h >> 16;
  v[2] = z.h & 0xFFFF;
  v[1] = z.l >> 16;
  v[0] = z.l & 0xFFFF;
  /* determine the number of significand places in the divisor */
  for (n = 4; v[n - 1] == 0; n--) ;
  /* normalize the divisor */
  vh = v[n - 1];
  for (d = 0; vh < 0x8000; d++) {
    vh <<= 1;
  }
  k = 0;
  for (j = 0; j < n + 4; j++) {
    t = (u[j] << d) + k;
    u[j] = t & 0xFFFF;
    k = t >> 16;
  }
  k = 0;
  for (j = 0; j < n; j++) {
    t = (v[j] << d) + k;
    v[j] = t & 0xFFFF;
    k = t >> 16;
  }
  vh = v[n - 1];
  vmh = (n > 1 ? v[n - 2] : 0);
  /* compute the quotient digits */
  for (j = 3; j >= 0; j--) {
    /* compute the quotient digit q[j] */
    /* first, find the trial quotient, qhat */
    t = (u[j + n] << 16) + u[j + n - 1];
    qhat = t / vh;
    rhat = t - vh * qhat;
    while (qhat == 0x10000 || qhat * vmh > (rhat << 16) + u[j + n - 2]) {
      qhat--;
      rhat += vh;
      if (rhat >= 0x10000) {
        break;
      }
    }
    /* then subtract ((qhat * v) << j) from u */
    k = 0;
    for (i = 0; i < n; i++) {
      t = u[i + j] + 0xFFFF0000 - k - qhat * v[i];
      u[i + j] = t & 0xFFFF;
      k = 0xFFFF - (t >> 16);
    }
    /* if the result was negative, decrease qhat by 1 */
    if (u[j + n] != k) {
      qhat--;
      k = 0;
      for (i = 0; i < n; i++) {
        t = u[i + j] + v[i] + k;
        u[i + j] = t & 0xFFFF;
        k = t >> 16;
      }
    }
    /* finally, store the quotient digit */
    q[j] = qhat;
  }
  /* unnormalize the remainder */
  mask = (1 << d) - 1;
  for (j = 3; j >= n; j--) {
    u[j] = 0;
  }
  k = 0;
  for (; j >= 0; j--) {
    t = (k << 16) + u[j];
    u[j] = t >> d;
    k = t & mask;
  }
  /* pack q and u to accumulator and arithAuxOut */
  accumulator.h = (q[3] << 16) + q[2];
  accumulator.l = (q[1] << 16) + q[0];
  arithAuxOut.h = (u[3] << 16) + u[2];
  arithAuxOut.l = (u[1] << 16) + u[0];
  return accumulator;
}


octa sdivOcta(octa x, octa y) {
  int sx, sy;
  octa xx, yy;
  octa q;

  if ((x.h & MSB(32)) != 0) {
    sx = 2;
    xx = subOcta(zero, x);
  } else {
    sx = 0;
    xx = x;
  }
  if ((y.h & MSB(32)) != 0) {
    sy = 1;
    yy = subOcta(zero, y);
  } else {
    sy = 0;
    yy = y;
  }
  q = udivOcta(zero, xx, yy);
  arithOvrflw = FALSE;
  switch (sx + sy) {
    case 0:
      /* +/+ */
      break;
    case 1:
      /* +/- */
      if (arithAuxOut.h != 0 || arithAuxOut.l != 0) {
        arithAuxOut = subOcta(arithAuxOut, yy);
      }
      /* negate q */
      if (arithAuxOut.h != 0 || arithAuxOut.l != 0) {
        q = subOcta(negone, q);
      } else {
        q = subOcta(zero, q);
      }
      break;
    case 2:
      /* -/+ */
      if (arithAuxOut.h != 0 || arithAuxOut.l != 0) {
        arithAuxOut = subOcta(yy, arithAuxOut);
      }
      /* negate q */
      if (arithAuxOut.h != 0 || arithAuxOut.l != 0) {
        q = subOcta(negone, q);
      } else {
        q = subOcta(zero, q);
      }
      break;
    case 3:
      /* -/- */
      arithAuxOut = subOcta(zero, arithAuxOut);
      if (q.h == MSB(32)) {
        arithOvrflw = TRUE;
      }
      break;
  }
  return q;
}


octa orOcta(octa x, octa y) {
  octa r;

  r.h = x.h | y.h;
  r.l = x.l | y.l;
  return r;
}


octa ornOcta(octa x, octa y) {
  octa r;

  r.h = x.h | ~y.h;
  r.l = x.l | ~y.l;
  return r;
}


octa norOcta(octa x, octa y) {
  octa r;

  r.h = ~(x.h | y.h);
  r.l = ~(x.l | y.l);
  return r;
}


octa xorOcta(octa x, octa y) {
  octa r;

  r.h = x.h ^ y.h;
  r.l = x.l ^ y.l;
  return r;
}


octa andOcta(octa x, octa y) {
  octa r;

  r.h = x.h & y.h;
  r.l = x.l & y.l;
  return r;
}


octa andnOcta(octa x, octa y) {
  octa r;

  r.h = x.h & ~y.h;
  r.l = x.l & ~y.l;
  return r;
}


octa nandOcta(octa x, octa y) {
  octa r;

  r.h = ~(x.h & y.h);
  r.l = ~(x.l & y.l);
  return r;
}


octa nxorOcta(octa x, octa y) {
  octa r;

  r.h = ~(x.h ^ y.h);
  r.l = ~(x.l ^ y.l);
  return r;
}


octa shlOcta(octa x, int n) {
  while (n >= 32) {
    x.h = x.l;
    x.l = 0;
    n -= 32;
  }
  if (n != 0) {
    x.h = (x.h << n) | (x.l >> (32 - n));
    x.l <<= n;
  }
  return x;
}


octa shrOcta(octa x, int n) {
  while (n >= 32) {
    x.l = x.h;
    x.h = 0;
    n -= 32;
  }
  if (n != 0) {
    x.l = (x.h << (32 - n)) | (x.l >> n);
    x.h >>= n;
  }
  return x;
}


octa sarOcta(octa x, int n) {
  while (n >= 32) {
    x.l = x.h;
    x.h = -(x.h >> 31);
    n -= 32;
  }
  if (n != 0) {
    x.l = (x.h << (32 - n)) | (x.l >> n);
    x.h = ((-(x.h >> 31)) << (32 - n)) | (x.h >> n);
  }
  return x;
}
