/*
 * common.h -- common definitions
 */


#ifndef _COMMON_H_
#define _COMMON_H_


#define VERSION		1		/* version of the MMIX architecture */
#define SUBVERSION	0		/* secondary byte of version number */
#define SUBSUBVERSION	2		/* further qualification of version */

#define LOG2_GR		8		/* log2 of number of global regs */
#define GR_NUM		(1 << LOG2_GR)	/* number of global registers */
#define GR_MASK		(GR_NUM - 1)	/* mask for global registers */

#define LOG2_LR		9		/* log2 of number of local regs */
#define LR_NUM		(1 << LOG2_LR)	/* number of local registers */
#define LR_MASK		(LR_NUM - 1)	/* mask for local registers */

#define MEMORY_SIZE	0x00100000	/* mem size, low part; high part = 0 */
#define ROM_SIZE	0x00001000	/* ROM size, low part; high part = 0 */

#define ROM_BASE_H	0x0000FFFF	/* ROM phys. base address, high part */
#define ROM_BASE_L	0x00000000	/* ROM phys. base address, low part */

#define IO_CLASS_MASK	0x7FFF0000
#define IO_NUMBER_MASK	0x0000FFFF

#define TERM_CLASS	0x00010000
#define DISK_CLASS	0x00020000

#define INSTRS_PER_MSEC	2500		/* average execution speed */

#define r_BIT		(1 << 7)	/* load without read permission */
#define w_BIT		(1 << 6)	/* store without write permission */
#define x_BIT		(1 << 5)	/* execute without exec permission */
#define n_BIT		(1 << 4)	/* instr refers to a negative addr */
#define k_BIT		(1 << 3)	/* instr is privileged (kernel) */
#define b_BIT		(1 << 2)	/* instr breaks the rules of GIMMIX */
#define s_BIT		(1 << 1)	/* instr violates security */
#define p_BIT		(1 << 0)	/* instr comes from a negative addr */

#define MASK(n)		((1 << (n)) - 1)
#define MSB(n)		(1 << ((n) - 1))

#define ZEXT(x, n)	((x) & MASK(n))
#define SEXT(x, n)	((x) & MSB(n) ? (x) | ~MASK(n) : (x) & MASK(n))


typedef int bool;			/* truth values */

#define FALSE		0
#define TRUE		1


typedef unsigned char byte;		/* 8 bit quantities */

typedef unsigned short wyde;		/* 16 bit quantities */

typedef unsigned int tetra;		/* 32 bit quantities */

typedef struct {
  tetra h;
  tetra l;
} octa;					/* 64 bit quantities */


#endif /* _COMMON_H_ */
