/*
 * wrtmbr.c -- write the master boot record to the disk
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>


#define SECTOR_SIZE	512
#define LINE_SIZE	100


void error(char *fmt, ...) {
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "Error: ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
  exit(1);
}


char getChar(FILE *bootFile) {
  static char line[LINE_SIZE];
  static int n = 0;
  static int i = 0;

  while (i == n) {
    if (fgets(line, LINE_SIZE, bootFile) == NULL) {
      error("unexpected EOF in boot sector file");
    }
    n = strlen(line);
    if (n == 0 || line[0] != ' ') {
      i = n;
    } else {
      i = 0;
    }
  }
  return line[i++];
}


unsigned char getNybble(FILE *bootFile) {
  int c;

  while (1) {
    c = getChar(bootFile);
    if (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
      continue;
    }
    if (c >= '0' && c <= '9') {
      return c - '0';
    }
    if (c >= 'A' && c <= 'F') {
      return c - 'A' + 10;
    }
    if (c >= 'a' && c <= 'f') {
      return c - 'a' + 10;
    }
    error("illegal character in boot sector file");
  }
}


unsigned char getByte(FILE *bootFile) {
  unsigned char b;

  b = getNybble(bootFile) << 4;
  b |= getNybble(bootFile);
  return b;
}


void usage(void) {
  fprintf(stderr, "Usage: wrtmbr <disk file> <master boot record file>\n");
  exit(1);
}


int main(int argc, char *argv[]) {
  FILE *diskFile;
  FILE *bootFile;
  unsigned char sectorBuffer[SECTOR_SIZE];
  int i;

  if (argc != 3) {
    usage();
  }
  diskFile = fopen(argv[1], "r+b");
  if (diskFile == NULL) {
    error("cannot open disk file '%s'", argv[1]);
  }
  bootFile = fopen(argv[2], "rt");
  if (bootFile == NULL) {
    error("cannot open boot sector file '%s'", argv[2]);
  }
  for (i = 0; i < SECTOR_SIZE; i++) {
    sectorBuffer[i] = getByte(bootFile);
  }
  fclose(bootFile);
  if (fseek(diskFile, 0, SEEK_SET) != 0) {
    error("cannot position to start of disk file");
  }
  if (fwrite(sectorBuffer, 1, SECTOR_SIZE, diskFile) != SECTOR_SIZE) {
    error("cannot write boot sector to disk file");
  }
  fclose(diskFile);
  return 0;
}
