/*
 * greg.c -- global register encoding
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "curslib.h"
#include "error.h"
#include "greg.h"


/*
 * These are the special global registers 0-31.
 * The table below needs no particular order, but
 * there must be no gaps in the register encoding.
 */
Greg gregTbl[32] = {
  { "rA",  rA,  "arithmetic status" },
  { "rB",  rB,  "bootstrap (trip)" },
  { "rC",  rC,  "cycle counter" },
  { "rD",  rD,  "dividend" },
  { "rE",  rE,  "epsilon" },
  { "rF",  rF,  "failure location" },
  { "rG",  rG,  "global threshold" },
  { "rH",  rH,  "himult" },
  { "rI",  rI,  "interval counter" },
  { "rJ",  rJ,  "return-jump" },
  { "rK",  rK,  "interrupt mask" },
  { "rL",  rL,  "local threshold" },
  { "rM",  rM,  "multiplex mask" },
  { "rN",  rN,  "serial number" },
  { "rO",  rO,  "register stack offset" },
  { "rP",  rP,  "prediction" },
  { "rQ",  rQ,  "interrupt request" },
  { "rR",  rR,  "remainder" },
  { "rS",  rS,  "register stack pointer" },
  { "rT",  rT,  "trap address" },
  { "rU",  rU,  "usage counter" },
  { "rV",  rV,  "virtual translation" },
  { "rW",  rW,  "where-interrupted (trip)" },
  { "rX",  rX,  "execution (trip)" },
  { "rY",  rY,  "Y operand (trip)" },
  { "rZ",  rZ,  "Z operand (trip)" },
  { "rBB", rBB, "bootstrap (trap)" },
  { "rTT", rTT, "dynamic trap address" },
  { "rWW", rWW, "where-interrupted (trap)" },
  { "rXX", rXX, "execution (trap)" },
  { "rYY", rYY, "Y operand (trap)" },
  { "rZZ", rZZ, "Z operand (trap)" }
};


Greg *gregCodeTbl[32];


static int gregCompare(const void *greg1, const void *greg2) {
  return strcmp(((Greg *) greg1)->name, ((Greg *) greg2)->name);
}


void initGregTable(void) {
  int i;

  /* first sort global register table alphabetically */
  qsort(gregTbl, sizeof(gregTbl)/sizeof(gregTbl[0]),
        sizeof(gregTbl[0]), gregCompare);
  /* then initialize global register code table */
  for (i = 0; i < 32; i++) {
    gregCodeTbl[i] = NULL;
  }
  for (i = 0; i < sizeof(gregTbl)/sizeof(gregTbl[0]); i++) {
    if (gregCodeTbl[gregTbl[i].code] != NULL) {
      error("global register table [%d] doubly initialized", i);
    }
    gregCodeTbl[gregTbl[i].code] = &gregTbl[i];
  }
  for (i = 0; i < 32; i++) {
    if (gregCodeTbl[i] == NULL) {
      error("global register table [%d] not initialized", i);
    }
  }
}


Greg *lookupGreg(char *name) {
  int lo, hi, tst;
  int res;

  lo = 0;
  hi = sizeof(gregTbl) / sizeof(gregTbl[0]);
  while (lo <= hi) {
    tst = (lo + hi) / 2;
    res = strcmp(gregTbl[tst].name, name);
    if (res == 0) {
      return &gregTbl[tst];
    }
    if (res < 0) {
      lo = tst + 1;
    } else {
      hi = tst - 1;
    }
  }
  return NULL;
}
