/*
 * mmu.c -- MMU simulation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "curslib.h"
#include "error.h"
#include "instr.h"
#include "cpu.h"
#include "mmu.h"
#include "memory.h"


//#define MMU_DEBUG


#ifdef MMU_DEBUG
#define DEBUG_VIRT(x)	cPrintf("DEBUG: virt = 0x%08X%08X\n", x.h, x.l);
#define DEBUG_PHYS(x)	cPrintf("DEBUG: phys = 0x%08X%08X\n", x.h, x.l);
#else /* MMU_DEBUG */
#define DEBUG_VIRT(x)
#define DEBUG_PHYS(x)
#endif /* MMU_DEBUG */


static octa v2p(octa vAddr, int writeAccess) {
  octa pAddr;

  if (vAddr.h & MSB(32)) {
    /* negative addresses are mapped directly to physical addresses */
    pAddr.h = vAddr.h & ~MSB(32);
    pAddr.l = vAddr.l;
    return pAddr;
  }
  /* otherwise, no address translation yet */
  return vAddr;
}


octa mmuRead64(octa vAddr) {
  octa pAddr;

  if (vAddr.l & 0x00000007) {
    throwException(ADDRESS_EXCEPTION);
  }
  DEBUG_VIRT(vAddr);
  pAddr = v2p(vAddr, 0);
  DEBUG_PHYS(pAddr);
  return memoryRead64(pAddr);
}


tetra mmuRead32(octa vAddr) {
  octa pAddr;

  if (vAddr.l & 0x00000003) {
    throwException(ADDRESS_EXCEPTION);
  }
  DEBUG_VIRT(vAddr);
  pAddr = v2p(vAddr, 0);
  DEBUG_PHYS(pAddr);
  return memoryRead32(pAddr);
}


wyde mmuRead16(octa vAddr) {
  octa pAddr;

  if (vAddr.l & 0x00000001) {
    throwException(ADDRESS_EXCEPTION);
  }
  DEBUG_VIRT(vAddr);
  pAddr = v2p(vAddr, 0);
  DEBUG_PHYS(pAddr);
  return memoryRead16(pAddr);
}


byte mmuRead8(octa vAddr) {
  octa pAddr;

  DEBUG_VIRT(vAddr);
  pAddr = v2p(vAddr, 0);
  DEBUG_PHYS(pAddr);
  return memoryRead8(pAddr);
}


void mmuWrite64(octa vAddr, octa data) {
  octa pAddr;

  if (vAddr.l & 0x00000007) {
    throwException(ADDRESS_EXCEPTION);
  }
  DEBUG_VIRT(vAddr);
  pAddr = v2p(vAddr, 1);
  DEBUG_PHYS(pAddr);
  memoryWrite64(pAddr, data);
}


void mmuWrite32(octa vAddr, tetra data) {
  octa pAddr;

  if (vAddr.l & 0x00000003) {
    throwException(ADDRESS_EXCEPTION);
  }
  DEBUG_VIRT(vAddr);
  pAddr = v2p(vAddr, 1);
  DEBUG_PHYS(pAddr);
  memoryWrite32(pAddr, data);
}


void mmuWrite16(octa vAddr, wyde data) {
  octa pAddr;

  if (vAddr.l & 0x00000001) {
    throwException(ADDRESS_EXCEPTION);
  }
  DEBUG_VIRT(vAddr);
  pAddr = v2p(vAddr, 1);
  DEBUG_PHYS(pAddr);
  memoryWrite16(pAddr, data);
}


void mmuWrite8(octa vAddr, byte data) {
  octa pAddr;

  DEBUG_VIRT(vAddr);
  pAddr = v2p(vAddr, 1);
  DEBUG_PHYS(pAddr);
  memoryWrite8(pAddr, data);
}


void mmuReset(void) {
  cPrintf("Resetting MMU...\n");
}


void mmuInit(void) {
  mmuReset();
}


void mmuExit(void) {
}
