/*
 * burn.c -- burn a new ROM (to C source, gets included into the simulator)
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "../../common.h"


#define WARNING	"/* THIS FILE IS GENERATED AUTOMATICALLY -- DO NOT EDIT */"

#define BUFFER_SIZE	100


void error(char *fmt, ...) {
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "Error: ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
  exit(1);
}


octa incrOcta(octa o, int delta) {
  octa r;

  r.h = o.h;
  r.l = o.l + delta;
  if (delta >= 0) {
    if (r.l < o.l) {
      r.h++;
    }
  } else {
    if (r.l > o.l) {
      r.h--;
    }
  }
  return r;
}


void byteOut(FILE *stream, byte b) {
  static int first = 1;
  static int n = 0;

  if (first) {
    first = 0;
  } else {
    if (n == 8) {
      n = 0;
      fprintf(stream, ",\n");
    } else {
      fprintf(stream, ", ");
    }
  }
  fprintf(stream, "0x%02X", b);
  n++;
}


void octaOut(FILE *stream, octa o) {
  int i;

  for (i = 0; i < 4; i++) {
    byteOut(stream, o.h >> (8 * (3 - i)));
  }
  for (i = 0; i < 4; i++) {
    byteOut(stream, o.l >> (8 * (3 - i)));
  }
}


void usage(void) {
  fprintf(stderr, "Usage: burn <hex file> <ROM file>\n");
  exit(1);
}


int main(int argc, char *argv[]) {
  FILE *hexFile;
  FILE *romFile;
  int line;
  char buffer[BUFFER_SIZE];
  octa currLoc, newLoc;
  octa currDat;

  if (argc != 3) {
    usage();
  }
  hexFile = fopen(argv[1], "rt");
  if (hexFile == NULL) {
    error("cannot open file '%s' for read", argv[1]);
  }
  romFile = fopen(argv[2], "wt");
  if (romFile == NULL) {
    error("cannot open file '%s' for write", argv[2]);
  }
  fprintf(romFile, "%s\n", WARNING);
  line = 0;
  currLoc.h = ROM_BASE_H;
  currLoc.l = ROM_BASE_L;
  while (1) {
    if (fgets(buffer, BUFFER_SIZE, hexFile) == NULL) {
      break;
    }
    line++;
    if (buffer[strlen(buffer) - 1] != '\n') {
      error("hex file line %d too long", line);
    }
    if (buffer[12] == ':') {
      /* set current location */
      if (sscanf(buffer, "%4x%8x", &newLoc.h, &newLoc.l) != 2) {
        error("improper hex file location in line %d", line);
      }
      newLoc.l &= 0xFFFFFFF8;
      currDat.h = 0;
      currDat.l = 0;
      while (currLoc.h != newLoc.h || currLoc.l != newLoc.l) {
        octaOut(romFile, currDat);
        currLoc = incrOcta(currLoc, 8);
      }
    } else
    if (buffer[0] == ' ') {
      /* read octabyte and advance current location */
      if (sscanf(buffer + 1, "%8x%8x", &currDat.h, &currDat.l) != 2) {
        error("improper hex file data in line %d", line);
      }
      octaOut(romFile, currDat);
      currLoc = incrOcta(currLoc, 8);
    } else {
      error("improper hex file line %d", line);
    }
  }
  fclose(hexFile);
  fclose(romFile);
  return 0;
}
